import json
import unittest
import os
import keyword
import re

try:
    with open("test.json") as f:
        test = json.load(f)
    lst = []
    for k in test['tasks'].keys():
        lst += list(test['tasks'][k].keys())
except:
    lst = []


class TestPyTest(unittest.TestCase):
    def test_json(self):
        self.assertTrue(os.path.isfile("test.json"),
                        "Missing file: test.json")
        self.assertIsInstance(test, dict,
                              "Json must be a dictionary")
        d = {"Json": [("configuration", dict), ("options", dict), ("tasks", dict)],
             "configuration": [("title", str), ("duration", int), ("teacher email address", str), ("bypass password", str), ("email user & password", list)],
             "options": [("send email", bool), ("python help()", bool), ("timer", bool)]}
        for key, value in d.items():
            for object_, class_ in value:
                self.assertIn(object_, list(test.get(key, test).keys()),
                              f"Missing key in {key}: {object_}")
                self.assertIsInstance(test.get(key, test)[object_], class_,
                                      f"{object_} must be from {class_}.")
        got_expected = {}
        for k in test['tasks'].keys():
            self.assertIsInstance(test['tasks'][k], dict,
                                  f"{k} must be a dictionary.")
            self.assertTrue(test['tasks'][k],
                            f"No tests for {k}")
            got_expected.update(test['tasks'][k])
        for name in lst:
            self.assertTrue(name.isidentifier() and not keyword.iskeyword(name),
                            f"Invalid name of function: '{name}'")
            self.assertTrue(lst.count(name) == 1,
                            f"The function '{name}' is repeated twice or more.")
        for k, v in got_expected.items():
            self.assertTrue(v,
                            f"No definitions for: {k}")
            self.assertIsInstance(v, list,
                                  f"{k} must be a list.")
            for x in v:
                self.assertIsInstance(x, list,
                                      f"Not all tests are lists in: {k}")
                self.assertTrue(len(x) > 1,
                                f"Error in task '{k}': {x} must be at format: [*parameters, return]")
        self.assertRegex(test['configuration']['teacher email address'], "(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)",
                         "Email address is invalid.")

    def test_txt(self):
        self.assertTrue(os.path.isfile("tasks.txt"),
                        "Missing file: tasks.txt")
        with open("tasks.txt", "r") as f:
            x = f.read()
        tasks = dict(
                zip([f.strip("# ") for f in re.findall("# .* #", x)],
                    [f.strip("\n") for f in re.split("# .* #", x)[1:]]))
        for t in lst:
            self.assertIn(t, tasks.keys(),
                          f"Task '{t}' has no instructions.")


if __name__ == '__main__':
    unittest.main(verbosity=2)
