import multiprocessing
import smtplib
from email.message import EmailMessage
import os
import io
from contextlib import redirect_stdout
import traceback
from tkinter import *
from tkinter.ttk import Scrollbar, Progressbar
import tkinter.messagebox as tkMessageBox
import tkinter.filedialog as tkFileDialog
from idlelib import macosx
from idlelib.multicall import MultiCallCreator
from idlelib import window
from idlelib import textview
from idlelib.editor import EditorWindow, get_accelerator
import json
import inspect
from idlelib.query import Query
import time
from datetime import datetime


__author__ = "Tom Ben Dor <tombendor.school@gmail.com>"
__version__ = "3.0.0"
__vars__ = {'__builtins__': __builtins__}

path = os.path.dirname(os.path.abspath(__file__))


with open(f"{path}/test.json") as f:
    test = json.load(f)

with open(f"{path}/tasks.txt", "r") as f:
    x = f.read()

tasks = dict(zip([f.strip("# ") for f in re.findall("# .* #", x)], [f.strip("\n") for f in re.split("# .* #", x)[1:]]))

EmailUserPass = test['configuration']['email user & password']


class TestWindow(EditorWindow):
    from idlelib.percolator import Percolator
    from idlelib.colorizer import ColorDelegator
    from idlelib.undo import UndoDelegator
    from idlelib.autoexpand import AutoExpand
    from idlelib.parenmatch import ParenMatch

    def __init__(self, root, test, tasks, clock=None):
        self.submenu = {}
        self.title = test['configuration']['title']
        self.test = test
        self.tasks = tasks
        self.clock = clock
        self.current = None
        self.list_of_tasks = lst = []
        self.got_expected = {}
        for k in self.test['tasks'].keys():
            lst += list(self.test['tasks'][k].keys())
            self.got_expected.update(self.test['tasks'][k])
        self.drafts = dict(zip(lst, map(lambda x: f"def {x}():\n\tpass\n", lst)))
        self.Ok = dict(zip(lst, [0] * len(lst)))
        self.opened_tasks = set()
        self.Query = Query
        self.menudefs = [
            ('edit', [
                ('_Undo', '<<undo>>'),
                ('_Redo', '<<redo>>'),
                None,
                ('Cu_t', '<<cut>>'),
                ('_Copy', '<<copy>>'),
                ('_Paste', '<<paste>>'),
                ('Select _All', '<<select-all>>'),
            ]),
            ('current task', [
                ('Run test', '<<run-test>>'),
                ("Check function", "<<check-func>>"),
                None,
                ('Show info', "<<show-task-info>>"),
            ]),
            ('help', [
                ('Instructions', '<<instructions>>'),
                ('Python help()', '<<python-help>>')
            ]),
            ('my test', [
                None,
                ('Show status', "<<show-status>>"),
                ('Show test', "<<show-test>>"),
                None,
                ('Submit test', "<<end-test>>")
            ]),
        ]
        self.keydefs = {
            '<<copy>>': ['<Control-c>', '<Control-C>'],
            '<<cut>>': ['<Control-x>', '<Control-X>'],
            '<<paste>>': ['<Control-v>', '<Control-V>'],
            '<<center-insert>>': ['<Control-l>'],
            '<<close>>': ['<Alt-F4>'],
            '<<do-nothing>>': ['<Control-x>'],
            '<<interrupt-execution>>': ['<Control-c>'],
            '<<open-class-browser>>': ['<Alt-c>'],
            '<<plain-newline-and-indent>>': ['<Control-j>'],
            '<<redo>>': ['<Control-y>'],
            '<<remove-selection>>': ['<Escape>'],
            '<<select-all>>': ['<Control-a>'],
            '<<toggle-auto-coloring>>': ['<Control-slash>'],
            '<<undo>>': ['<Control-z>'],
            '<<smart-backspace>>': ['<Key-BackSpace>'],
            '<<newline-and-indent>>': ['<Key-Return>', '<Key-KP_Enter>'],
            '<<smart-indent>>': ['<Key-Tab>'],
            '<<indent-region>>': ['<Control-Key-bracketright>'],
            '<<dedent-region>>': ['<Control-Key-bracketleft>'],
            '<<del-word-left>>': ['<Control-Key-BackSpace>'],
            '<<del-word-right>>': ['<Control-Key-Delete>'],
            '<<force-open-completions>>': ['<Control-Key-space>'],
            '<<expand-word>>': ['<Alt-Key-slash>'],
            '<<force-open-calltip>>': ['<Control-Key-backslash>'],
            '<<show-status>>': ['<Control-r>'],
            '<<run-test>>': ['<F5>'],
            '<<instructions>>': ['<F1>'],
            '<<bypass>>': ['<Alt-b>']
        }
        self.root = root
        self.menubar = Menu(self.root)
        self.top = top = window.ListedToplevel(self.root, menu=self.menubar)

        self.text_frame = text_frame = Frame(top)
        self.vbar = vbar = Scrollbar(text_frame, name='vbar')

        text_options = {
            'name': 'text',
            'padx': 5,
            'pady': 0,
            'wrap': 'none',
            'width': 80,
            'tabstyle': 'wordprocessor',  # new in 8.5
            'height': 25,
        }
        self.text = text = MultiCallCreator(Text)(text_frame, **text_options)
        self.top.focused_widget = self.text
        self.createmenubar()
        text.bind("<<cut>>", self.cut)
        text.bind("<<copy>>", self.copy)
        text.bind("<<paste>>", self.paste)
        text.bind("<<do-nothing>>", lambda event: "break")
        text.bind("<<select-all>>", self.select_all)
        text.bind("<<remove-selection>>", self.remove_selection)
        text.bind("<<smart-backspace>>", self.smart_backspace_event)
        text.bind("<<newline-and-indent>>", self.newline_and_indent_event)
        text.bind("<<smart-indent>>", self.smart_indent_event)
        parenmatch = self.ParenMatch(self)
        text.bind("<<flash-paren>>", parenmatch.flash_paren_event)
        text.bind("<<paren-closed>>", parenmatch.paren_closed_event)
        text.bind("<<bypass>>", lambda x: self.bypass())
        vbar['command'] = self.handle_yview
        vbar.pack(side=RIGHT, fill=Y)
        text['yscrollcommand'] = vbar.set
        text_frame.pack(side=LEFT, fill=BOTH, expand=1)
        text.pack(side=TOP, fill=BOTH, expand=1)
        text.focus_set()

        self.usetabs = True
        self.tabwidth = 4
        self.indentwidth = 4
        self.text['tabs'] = 32

        self.context_use_ps1 = False

        self.num_context_lines = 50, 500, 5000000
        self.per = per = self.Percolator(text)
        self.undo = undo = self.UndoDelegator()
        per.insertfilter(undo)
        text.undo_block_start = undo.undo_block_start
        text.undo_block_stop = undo.undo_block_stop
        self.color = None

        text.event_add('<<autocomplete>>', '<Key-Tab>')
        text.event_add('<<try-open-completions>>', '<KeyRelease-period>',
                       '<KeyRelease-slash>', '<KeyRelease-backslash>')
        text.event_add('<<try-open-calltip>>', '<KeyRelease-parenleft>')
        text.event_add('<<refresh-calltip>>', '<KeyRelease-parenright>')
        text.event_add('<<paren-closed>>', '<KeyRelease-parenright>',
                       '<KeyRelease-bracketright>', '<KeyRelease-braceright>')

        text.bind("<<expand-word>>", self.AutoExpand(self).expand_word_event)
        text.bind("<<end-test>>", self.end_test)
        text.bind("<<run-test>>", lambda x: self.run_test())
        text.bind("<<show-status>>", lambda x: self.show_result(show=True))
        text.bind("<<python-help>>", lambda x: self.python_help())
        text.bind("<<show-task-info>>", lambda x: self.show_task_file())
        text.bind("<<check-func>>", self.check_func)
        text.bind("<<instructions>>", self.show_instructions)
        text.bind("<<show-test>>", self.show_test)
        self.menubar.entryconfigure("Current task", state=DISABLED)
        self.menubar.entryconfigure("Edit", state=DISABLED)

        self.top.attributes("-topmost", True)
        self.top.protocol("WM_DELETE_WINDOW", self.close)
        self.top.resizable(0, 0)

        self.text['font'] = ('Calibri', 12)
        self.top.title(f"{test['configuration']['title']}")
        self.name = name = GetName(title="PyTest", message="Please enter your full name here:",
                                   parent=self.top, ).result
        if not name:
            exit()
        self.top.title(f"{test['configuration']['title']} - {name}")

        self.timer = isinstance(clock, Label)
        if not self.timer:
            if not GetPass(title="No timer!",
                           message="In order to take a test without timer,\nplease enter password:",
                           parent=self.text).result:
                exit()

        welcome = f"Hello {self.name.split(' ')[0]}! Welcome to PyTest!\n\n" \
                  "For instructions, please enter F1.\n\n" \
                  "Please go to 'My test' menu and pick a task to start with.\n\n"
        text.config(state=NORMAL)
        text.insert(1.0, welcome)
        text.config(state=DISABLED)

        self.apply_bindings()
        self.menudict['help'].entryconfig('Python help()',
                                          state=NORMAL if test['options']['python help()'] else DISABLED)

        self.start_time = get_time()
        self.bypasses = 0
        self.email_sent = False
        self.timer_status = "On"
        if self.timer:
            self.create_progress_bar()
        self.prompt_last_line = ''

    def create_progress_bar(self):
        start = time.time()
        pb = Progressbar(self.text_frame, length=250, orient=HORIZONTAL, mode="determinate",
                         maximum=self.clock.test_time)
        pb.pack(side=RIGHT)
        l = Label(self.text_frame, text="Test is on air...", font=("Calibri", 12))
        l.pack(side=LEFT)

        def bar():
            time_ = (time.time() - start)
            if time_ > self.clock.test_time:
                l.destroy()
                pb.destroy()
                self.timesup()
                return "break"
            pb['value'] = self.clock.test_time - time_
            pb['maximum'] = self.clock.test_time
            p = '{:.1%}'.format(sum(self.Ok.values())/sum(map(len, self.got_expected.values())))
            l.config(text=f"Test is on air  //  Progress currently on: {p}  //  Time left:")
            pb.after(10, bar)

        bar()

    def get_function_return(self, code=None, task=None, index=0):
        args = (code, task, tuple(self.got_expected[task][index][:-1]))
        d = multiprocessing.Process(target=exec_func, args=args)
        d.start()
        d.join(2)
        if d.is_alive():
            d.terminate()
            raise TookTooLong
        return exec_func(*args)

    def show_test(self, event=None):
        textview.view_text(self.text, f"{self.title} - {self.name}", self.get_test())

    def python_help(self):
        statement = self.Query(title="Python help", message="Enter the statement here:", parent=self.text).result
        if statement:
            textview.view_text(self.text, f"help({statement})", self.output(lambda file: help(statement)))

    def show_task_file(self):
        if self.current is None:
            return "break"
        if self.current in self.tasks.keys():
            textview.view_text(self.text, self.current, self.tasks[self.current])

    def task(self, func_name, topic_name):
        self._addcolorizer()
        self.text['font'] = ('courier', 10)
        self.text.config(state=NORMAL)
        self.menubar.entryconfigure("Current task", state=NORMAL)
        self.menubar.entryconfigure("Edit", state=NORMAL)
        if self.current != func_name:
            self.menudict['current task'].entryconfigure('Check function', state=DISABLED)
        self.save_draft()
        self.current = func_name
        self.current_topic = topic_name
        self.top.title(f"{self.title} - {self.name} - {self.current}")
        self.text.delete(1.0, END)
        code = self.drafts[self.current]
        self.text.insert(1.0, code)
        if not func_name in self.opened_tasks:
            self.show_task_file()
        self.undo.reset_undo()
        self.opened_tasks.add(func_name)

    def timesup(self):
        self.current = None
        if self.timer:
            if self.timer_status == "Off":
                return "break"
            self.timer_status = "Off"
            self.clock.tick = False
            self.clock.test_time = 0
            self.clock.config(text="PyTest", bg='white', font=('Gisha', 18, 'bold'))
        self.text['font'] = ('Calibri', 12)
        self._rmcolorizer()
        self.top.title(f"{self.title} - {self.name}")
        self.text.config(state=NORMAL)
        self.text.delete(1.0, END)
        score, details = self.show_result(show=False)
        self.text.insert(1.0,
                         f"Time's up!\n\nHere are your results:\n\nYour score is {score}\n\n{details}")
        self.text.config(state=DISABLED)
        self.menubar.entryconfigure("Edit", state=DISABLED)
        self.menubar.entryconfigure("Current task", state=DISABLED)
        self.menudict['help'].entryconfigure("Python help()", state=DISABLED)
        for k in self.test['tasks'].keys():
            self.menudict['my test'].entryconfigure(k, state=DISABLED)
        self.top.protocol("WM_DELETE_WINDOW", lambda: self.end_test(ask=False))
        self.menudict['my test'].entryconfigure("Submit test",
                                                command=lambda: self.end_test(ask=False))
        self.undo.reset_undo()
        tkMessageBox.showwarning("Time's up", "Your test is over.", parent=self.text)

    def kill(self):
        self.top.withdraw()
        tkMessageBox.showinfo("PyTest", "Thank you for using PyTest.", parent=self.text)
        self.root.destroy()

    def end_test(self, ask=True):
        if ask:
            if not tkMessageBox.askyesno(title="Submit test",
                                         message="Are you sure you want to submit your test?", parent=self.text,
                                         default="no"):
                return "break"
        if self.clock is not None:
            self.clock.tick = False
            self.clock.config(text="PyTest", bg='white', font=('Gisha', 18, 'bold'))
        if self.test['options']['send email']:
            self.send_email()
            if self.email_sent:
                self.kill()
            return "break"
        self.kill()

    def show_result(self, show=True):
        details = ""
        for k, v in self.Ok.items():
            details += f"{k} : {v} / {len(self.got_expected[k])}\n"
        score = f"{sum(self.Ok.values())} / {sum(map(len, self.got_expected.values()))}"
        if show:
            tkMessageBox.showinfo(f"Your results - {score}", details, parent=self.text)
        return score, details

    def email(self, Msg, admin=False):
        try:
            server = smtplib.SMTP_SSL('smtp.gmail.com', 465)
            server.login(*EmailUserPass)
            server.send_message(Msg)
        except Exception as e:
            if not admin:
                if tkMessageBox.askretrycancel("Can't send results email!",
                                               "Please check your internet connection and retry.",
                                               parent=self.text,
                                               default="retry"):
                    self.email(Msg)
                else:
                    self.email_error()
            else:
                if tkMessageBox.askretrycancel("ERROR!", str(e), parent=self.text, default="retry"):
                    self.email(Msg, admin=True)
        else:
            tkMessageBox.showinfo("Email was sent successfully.",
                                  "Your results were sent successfully.",
                                  parent=self.text)
            self.email_sent = True

    def send_email(self, admin=False, send=True):
        msg = EmailMessage()
        msg['Subject'] = f"{self.title} - {self.name} - Results"
        msg['To'] = self.test['configuration']['teacher email address']
        msg['from'] = "tombendor.spam@gmail.com"
        details = ""
        for k, v in self.Ok.items():
            details += f"{k} : {v} / {len(self.got_expected[k])}\n"
        msg.set_content(
                f"Here are {self.name.split(' ')[0]}'s results:\n\nOk: {sum(self.Ok.values())} / {sum(map(len, self.got_expected.values()))}\n\n{details}\n\n"
                f"Start time: {self.start_time}\nEnd time: {get_time()}\n\nBypasses: {self.bypasses}; Timer: {self.timer}")
        text = msg.get_content()
        msg.add_attachment(self.get_test().encode(), maintype="text", subtype="text", filename=f"{self.name}.txt")
        if send:
            self.email(msg, admin=admin)
        return text

    def email_error(self):
        self.text.config(state=NORMAL)
        self.text.delete(1.0, END)
        self.text.insert(1.0, "Failed to send email.\n\nBypass required.")
        self.text.config(state=DISABLED)
        self.top.title(f"{self.title} - {self.name}")
        self.menubar.entryconfig("My test", state=DISABLED)
        self.menubar.entryconfig("Help", state=DISABLED)

        def kill():
            if GetPass(title="Close window", message="Enter password for closing the window:", parent=self.text).result:
                self.kill()

        self.top.protocol("WM_DELETE_WINDOW", kill)

    def save_draft(self):
        if self.current is None:
            return "break"
        self.drafts[self.current] = self.text.get(1.0, END)

    @staticmethod
    def output(command, *args, **kwargs):
        with io.StringIO() as baf, redirect_stdout(baf):
            command(*args, **kwargs, file=baf)
            output = baf.getvalue()
        return output

    def run_test(self):
        if self.current is None or self.text['state'] == DISABLED:
            return "break"
        self.save_draft()
        self.menudict['current task'].entryconfigure('Check function', state=DISABLED)
        code = self.text.get(1.0, END)
        self.current_code = code
        code = code.split("\n")
        code = [l for l in code if
                l[:1] == "\t" or l.startswith("def ") or l.startswith("import ") or l.startswith("from ")]
        code = '\n'.join(code)

        class Error(Exception):
            pass

        try:
            code_ = compile(code, "file", mode="exec")
        except (SyntaxError, OverflowError, ValueError) as value:
            lineno = getattr(value, 'lineno', '') or 1
            offset = getattr(value, 'offset', '') or 0
            if offset == 0:
                lineno += 1
            pos = "0.0 + %d lines + %d chars" % (lineno - 1, offset - 1)
            self.colorize_syntax_error(self.text, pos)
            tkMessageBox.showerror("SyntaxError", "Fix your code and retry.", parent=self.text)
        else:
            vars_ = __vars__.copy()
            exec(code_, vars_, vars_)
            if self.current in vars_.keys() and callable(vars_[self.current]):
                func = vars_[self.current]
                self.current_func = func
                for k, v in vars_.items():
                    globals()[k] = v
                rslt, sumOk = "", 0
                index = 0
                for *p, e in self.got_expected[self.current]:
                    if count_parameters(func) != len(p):
                        tkMessageBox.showerror("Parameters error.",
                                               f"The function takes {count_parameters(func)} parameters instead of {len(p)}.",
                                               parent=self.text)
                        return "break"
                    try:
                        got = self.get_function_return(code=code, task=self.current, index=index)
                    except TookTooLong:
                        tkMessageBox.showerror("Took too long!",
                                               "Your code didn't respond for more then 2 seconds.\n"
                                               "Please check your for & while loops.",
                                               parent=self.text)
                        return "break"
                    except Exception as e_:
                        output = self.output(traceback.print_exc).split("\n")
                        for i, l in enumerate(output):
                            if "return eval(name, vars_, vars_)(*args)" in l:
                                output = [output[0]] + output[i + 1:]
                                break
                        output = '\n'.join(output)
                        output = output.strip().replace('File "<string>"', f'Task "{self.current}"')
                        output = f"Error while testing {self.current} with the following parameters: {parameter((p, ))}\n\n" + output
                        tkMessageBox.showerror(type(e_).__name__, output, parent=self.text)
                        got = Error(type(e_).__name__)
                    else:
                        lst = [str(parameters_got[-1]) for task in
                               [item for task in test['tasks'].values() for item in task.values()] for parameters_got in
                               task]
                        if len(str(got)) > len(max(lst, key=len)):
                            got = Error("Returned value is too large")
                    index += 1
                    if got == e:
                        rslt += "OK - "
                        sumOk += 1
                        o = True
                    else:
                        rslt += "X - "
                        o = False
                    if not isinstance(got, Error):
                        got = got if not isinstance(got, str) else "'" + got + "'"
                        rslt += f"{parameter(p)} - Got: {got}"
                    else:
                        rslt += f"{parameter(p)} - Error: {got}"
                        rslt += f"\n\n=== skipped  {len(self.got_expected[self.current]) - index} ==="
                        tkMessageBox.showinfo(f"{self.current} -  {sumOk} / {len(self.got_expected[self.current])}",
                                              rslt,
                                              parent=self.text)
                        return "break"
                    if not o:
                        rslt += f"\nExpected: '{e}'" if isinstance(e, str) else f"\nExpected: {e}"
                    rslt += "\n\n"
                tkMessageBox.showinfo(f"{self.current} -  {sumOk} / {len(self.got_expected[self.current])}", rslt,
                                      parent=self.text)
                if sumOk == len(self.got_expected[self.current]):
                    self.submit_task()
                    tkMessageBox.showinfo("Well done!",
                                          f"You completed {self.current}.\n\nContinue to the next task or submit your test.",
                                          parent=self.text)
                self.menudict['current task'].entryconfigure('Check function', state=NORMAL)
                self.Ok[self.current] = sumOk
                self.submenu[self.current_topic].entryconfig(self.current,
                                                             accelerator=f"[{sumOk} / {len(self.got_expected[self.current])}]")
                if sum(self.Ok.values())/sum(map(len, self.got_expected.values())) == 1:
                    tkMessageBox.showinfo("Test was completed!",
                                          "Great! you can now submit your test.\n"
                                          "Just go to the 'My test' section and click: 'Submit test'.",
                                          parent=self.text)
            else:
                tkMessageBox.showwarning("Function was not found",
                                         f"Define the following function: '{self.current}'", parent=self.text)

    def check_func(self, x):
        if self.text.get(1.0, END) != self.current_code:
            tkMessageBox.showerror("Code has changed!", "Please run new test before checking the function.",
                                   parent=self.text)
            return "break"
        func = self.current_func
        current = self.current
        query = self.Query

        class CheckFunc(query):
            def create_widgets(self):
                super().create_widgets()
                clear = self.button_cancel
                clear.config(text="Clear", command=self.clear)

            def clear(self):
                self.entry_error['text'] = ""
                self.entry.delete(0, END)

            def run(self, func, *args):
                self.entry_error['foreground'] = 'black'

                def f(*a):
                    return ', '.join([str(x) if not isinstance(x, str) else f"'{x}'" for x in a])

                try:
                    x = func(*args)
                except Exception as e:
                    self.showerror(f"{type(e).__name__}: {str(e)}")
                else:
                    self.entry_error['text'] = f"{current}({f(*args)}) = {x}"

            def showerror(self, x, widget=None):
                self.entry_error['foreground'] = 'red'
                self.entry_error['text'] = x

            def entry_ok(self):
                entry = self.entry.get().strip()
                vars_ = __vars__.copy()
                if entry:
                    try:
                        entry = eval(entry, vars_, vars_)
                    except Exception as e:
                        self.showerror("ERROR: " + str(e))
                        return None
                    else:
                        if isinstance(entry, tuple):
                            if len(entry) != count_parameters(func):
                                self.showerror(
                                        f"You entered {len(entry)} parameters instead of {count_parameters(func)}.")
                                return None
                            self.run(func, *entry)
                        else:
                            if count_parameters(func) != 1:
                                self.showerror("Missing some parameters.")
                                return None
                            self.run(func, entry)
                else:
                    self.showerror('ERROR: blank line.')

        CheckFunc(title="Check function", message="Enter your input here:", parent=self.text)

    def submit_task(self, state=DISABLED):
        self.text.config(state=state)
        self.submenu[self.current_topic].entryconfigure(self.current, state=state)
        self.menubar.entryconfigure("Current task", state=state)
        self.menubar.entryconfigure("Edit", state=state)

    def get_test(self):
        info = ""
        for k, v in self.drafts.items():
            if k not in self.opened_tasks:
                continue
            title = f"{k} - [{self.Ok[k]} / {len(self.got_expected[k])}]"
            v = v.strip("\n")
            info += f"{title}\n{'-' * len(title)}\n{v}\n\n"
        return info.replace("\t", "    ").strip("\n")

    def bypass(self, ask=True):
        self.menubar.delete(5)
        if ask:
            if not GetPass(title="Bypass", message="Enter password for bypass:", parent=self.text).result:
                return "break"
        bypass = Menu(self.menubar, tearoff=0)
        for k in self.test['tasks'].keys():
            m = Menu(bypass, tearoff=0)
            for v in self.test['tasks'][k].keys():
                b = Menu(m, tearoff=0)

                def func(k=k, v=v):
                    if self.submenu[k].entrycget(v, "state") == DISABLED:
                        self.submenu[k].entryconfig(v, state=NORMAL)
                        if v == self.current:
                            self.submit_task(state=NORMAL)
                    elif self.submenu[k].entrycget(v, "state") == NORMAL:
                        self.submenu[k].entryconfig(v, state=DISABLED)
                        if v == self.current:
                            self.submit_task()
                    self.bypass(ask=False)

                if self.submenu[k].entrycget(v, "state") == DISABLED:
                    b.add_command(label="Unlock task", command=func)
                if self.submenu[k].entrycget(v, "state") == NORMAL:
                    b.add_command(label="Lock task", command=func)

                def setOk(k=k, v=v):
                    class GetOk(self.Query):
                        def entry_ok(self):
                            self.entry_error['text'] = ''
                            entry = self.entry.get().strip()
                            try:
                                entry = int(entry)
                            except:
                                self.showerror('Not a valid number.')
                            else:
                                return entry

                    ok = GetOk(title="Set ok", message="Please enter a number:", parent=self.text).result
                    if ok is None:
                        return "break"
                    ok = max(0, ok)
                    ok = min(len(self.got_expected[v]), ok)
                    self.submenu[k].entryconfig(v, accelerator=f"[{ok} / {len(self.got_expected[v])}]")
                    self.Ok[v] = ok

                b.add_command(label="Set ok", command=setOk)
                m.add_cascade(label=v, menu=b)
            m.add_separator()

            def func(k=k):
                d = {DISABLED: NORMAL, NORMAL: DISABLED}
                self.menudict['my test'].entryconfig(k,
                                                     state=d[self.menudict['my test'].entrycget(k, "state")])
                self.bypass(ask=False)

            if self.menudict['my test'].entrycget(k, "state") == NORMAL:
                m.add_command(label="Lock topic", command=func)
            if self.menudict['my test'].entrycget(k, "state") == DISABLED:
                m.add_command(label="Unlock topic", command=func)
            bypass.add_cascade(label=k, menu=m)

        bypass.add_separator()
        em = Menu(bypass, tearoff=0)

        def toggle_email():
            self.test['options']['send email'] = not self.test['options']['send email']
            em.entryconfig("Send email after test", accelerator=str(self.test['options']['send email']))

        em.add_command(label="Send email after test", command=toggle_email,
                       accelerator=str(self.test['options']['send email']))
        em.add_command(label="Send email now", command=lambda: self.send_email(admin=True))

        bypass.add_cascade(label="Send email", menu=em)
        self.bypasses += 1

        def change_time():
            clock = self.clock

            class GetTime(self.Query):
                def entry_ok(self):
                    self.entry_error['text'] = ''
                    entry = self.entry.get().strip()
                    try:
                        entry = int(entry)
                    except:
                        self.showerror('Not a valid number.')
                    else:
                        if clock.test_time + entry * 60 >= 36000:
                            self.showerror('Time is too long.')
                            return None
                        if clock.test_time + entry * 60 <= 0:
                            self.showerror("Too less time.\nTake a look at the Time's up section.")
                            return None
                        return entry

            time = GetTime(title="Change time", message="Please enter the time change in minutes",
                           parent=self.text).result
            if time:
                self.clock.test_time += time * 60

        bypass.add_command(label="Add / reduce time", command=change_time,
                           state=NORMAL if self.timer and self.timer_status == "On" else DISABLED)

        def timesup():
            self.timesup()
            self.bypass(ask=False)

        bypass.add_command(label="Time's up", command=timesup,
                           state=NORMAL if self.timer_status == "On" else DISABLED)

        def set_name():
            self.menubar.delete(5)
            name = GetName(title="PyTest", message="Please enter your full name here:", parent=self.text,
                           text0=self.name).result
            if name is None:
                return "break"
            self.name = name
            self.top.title(f"{self.title} - {name + ((' - ' + self.current) if self.current else '')}")

        bypass.add_command(label="Set name", command=set_name, accelerator=self.name)

        def python_help():
            d = {DISABLED: NORMAL, NORMAL: DISABLED}
            self.menudict['help'].entryconfig('Python help()',
                                              state=d[self.menudict['help'].entrycget('Python help()', "state")])
            self.bypass(ask=False)

        bypass.add_command(label="Python help()", command=python_help,
                           accelerator=self.menudict['help'].entrycget('Python help()', "state"))
        bypass.add_command(label="Create .txt file", command=lambda: self.create_txt_file(show=True))
        bypass.add_command(label="Kill program", command=self.kill)
        bypass.add_separator()
        bypass.add_command(label="Exit bypass menu", command=lambda: self.menubar.delete(5))
        self.menubar.add_cascade(menu=bypass, label="Admin access")

    def create_txt_file(self, show=False):
        path_ = tkFileDialog.askdirectory()
        if os.path.isfile(f"{path_}/{self.title} - {self.name}.txt") and show:
            if not tkMessageBox.askyesno("File already in path.", "This will delete the current file.\nContinue?",
                                         parent=self.text):
                return "break"
        text = f"{self.send_email(send=False)}\n\n{self.get_test()}"
        with open(f"{path_}/{self.title} - {self.name}.txt", "w") as f:
            f.write(text)
        if show:
            tkMessageBox.showinfo("File was created.", f"txt file was created at path: {path_}", parent=self.text)

    def close(self):
        if not (self.opened_tasks or self.timer):
            self.kill()
            return "break"
        tkMessageBox.showerror(title="Test is still on air!",
                               message="For ending your test, please click 'Submit test'",
                               parent=self.text)

    menu_specs = [
        ("edit", "Edit"),
        ("current task", "Current task"),
        ("my test", "My test"),
        ("help", "Help")]

    def createmenubar(self):
        mbar = self.menubar
        self.menudict = menudict = {}
        for name, label in self.menu_specs:
            menudict[name] = menu = Menu(mbar, name=name, tearoff=0)
            mbar.add_cascade(label=label, menu=menu)
        submenu = self.submenu
        for k, v in list(self.test['tasks'].items())[::-1]:
            m = Menu(mbar, tearoff=0)
            for f in v:
                def command(f=f, k=k):
                    self.task(func_name=f, topic_name=k)

                m.add_command(label=f, command=command, accelerator=f"[0 / {len(self.got_expected[f])}]")
            submenu[k] = m
            self.menudict['my test'].insert_cascade(0, label=k, underline=0, menu=m)
        self.fill_menus()

    def _addcolorizer(self):
        if self.color:
            return "break"
        self.color = self.ColorDelegator()
        self.per.insertfilter(self.color)

    def _rmcolorizer(self):
        if not self.color:
            return "break"
        self.color.removecolors()
        self.per.removefilter(self.color)
        self.color = None

    def apply_bindings(self):
        text = self.text
        text.keydefs = self.keydefs
        for event, keylist in self.keydefs.items():
            if keylist:
                text.event_add(event, *keylist)

    def fill_menus(self):
        menudefs = self.menudefs
        menudict = self.menudict
        keydefs = self.keydefs
        text = self.text
        for mname, entrylist in menudefs:
            menu = menudict.get(mname)
            if not menu:
                continue
            for entry in entrylist:
                if not entry:
                    menu.add_separator()
                else:
                    label, eventname = entry
                    accelerator = get_accelerator(keydefs, eventname)

                    def command(text=text, eventname=eventname):
                        text.event_generate(eventname)

                    menu.add_command(label=label,
                                     command=command,
                                     accelerator=accelerator)

    def show_instructions(self, x):
        with open(f"{path}/Instructions.txt", "r") as i:
            textview.view_text(self.text_frame, "Instructions", i.read())

    def _make_blanks(self, n):
        ntabs = n // self.tabwidth
        return '\t' * ntabs


def exec_func(code, name, args):
    vars_ = __vars__.copy()
    exec(code, vars_, vars_)
    return eval(name, vars_, vars_)(*args)


class GetName(Query):
    def entry_ok(self):
        self.entry_error['text'] = ''
        entry = self.entry.get().strip().title()
        if not re.match("[A-Za-z ]{3,20}$", entry):
            self.showerror('Not a valid name!')
            return None
        return entry


class GetPass(Query):
    password = test['configuration']['bypass password']

    def create_widgets(self):
        super().create_widgets()
        self.entry.config(show="*")

    def entry_ok(self):
        self.entry_error['text'] = ''
        entry = self.entry.get()
        if entry != self.password:
            self.entry_error['text'] = "Incorrect password."
            return None

        def func():
            self.result = True
            self.destroy()

        self.entry_error.config(foreground='green', text="Correct password.")
        self.after(750, func)


class TookTooLong(Exception):
    pass


def parameter(*p):
    def f(x):
        if isinstance(x, str):
            return f"'{x}'"
        return str(x)

    p = p[0]
    if len(p) == 1:
        return p[0]
    return f"({', '.join([f(x) for x in p])})"


def count_parameters(func):
    p = inspect.signature(func)
    p = p.parameters
    return len(p)


def get_time():
    now = datetime.now()
    time = now.strftime("%B %d, %Y - %X")
    return time


def main():
    test_time = 60 * test['configuration']['duration']

    root = Tk()
    root.resizable(width=False, height=False)
    root.protocol("WM_DELETE_WINDOW", lambda: None)
    root.geometry("140x30+0+0")
    root.overrideredirect(1)
    root.attributes("-topmost", True)

    clock = Label(root, font=('Gisha', 18, 'bold'), bg='white', text="PyTest", borderwidth=2, relief="ridge")
    clock.tick = True
    clock.test_time = test_time
    clock.pack(fill=BOTH)

    def tick(countdown=True):
        if not clock.tick:
            return "break"

        if countdown:
            time_left = int(clock.test_time) - int(time.time() - start)
        else:
            time_left = int(time.time() - start)

        if time_left <= 0 and countdown:
            timesup()
            return "break"

        h = '{:02d}'.format((time_left // 60) // 60)
        m = '{:02d}'.format((time_left // 60) % 60)
        s = '{:02d}'.format(time_left % 60)

        clock.config(text=f"{h}:{m}:{s}", font=('Gisha', 24, 'bold'))
        if time_left / clock.test_time <= 0.25 and countdown:
            clock.config(bg='#f6a6b2')
        else:
            clock.config(bg='#b7ded2')
        clock.after(1000, lambda: tick(countdown))

        if h == 0 and m in [45, 30, 10, 5, 3] and s == 0 and countdown:
            tkMessageBox.showwarning(f"{m} minuets left", f"You have only {m} minuets left!", parent=app)

    def timesup():
        clock.tick = False
        clock.config(text="PyTest", bg='white', font=('Gisha', 18, 'bold'))

    app = Tk()
    app.withdraw()

    if test['options']['timer']:
        TestWindow(root=app, test=test, tasks=tasks, clock=clock)
        start = time.time()
        tick()
    else:
        TestWindow(root=app, test=test, tasks=tasks, clock=None)
        start = time.time()
        tick(False)
    mainloop()


if __name__ == '__main__':
    main()
